#!/usr/bin/python3
#
# Attempting a 15-puzzle solver
#
import  sys
import  random

bTest = False

class Pos:
    def __init__ (self, x, y):
        self.x = x
        self.y = y

def Parity (tiles):
    n = 0;
    for i in range (len(tiles)):
        for j in range (i+1, len(tiles)):
            if ( tiles[i] > tiles[j] ):
                n += 1
    return n

class Puzzle:
    def __init__ (self, argv):
        if (len (argv) > 1):
            with open (argv[1], 'r') as fIn:
                tiles = []
                for sLine in fIn:
                    sLine.strip ()
                    if ( sLine == '' ):
                        break
                    tiles.append (int (sLine))
            if ( Parity(tiles) & 1 ):
                print ("Parity Error")
                sys.exit (1)
        else:
            tiles = [i for i in range(1, 16)]
            random.shuffle (tiles)
            if ( Parity(tiles) & 1 ):
                i = tiles[-2]
                tiles[-2] = tiles[-1]
                tiles[-1] = i
            with open ('Fifteen.dat', 'w') as fOut:
                for cell in tiles:
                    fOut.write ('{:d}\n'.format (cell))
        tiles.append (0)
        self.grid = [tiles[0:4], tiles[4:8], tiles[8:12], tiles[12:16]]
        self.lock = [[False for j in range(4)] for i in range(4)]
        self.space = self.Find (0)

    def Print (self):
        symbol = " 123456789ABCDEF"
        print ()
        for row in self.grid:
            line = ''
            for cell in row:
                line += symbol[cell]
            print (line)
        if ( not bTest ):
            input ()

    def Find (self, n):
        for y, row in enumerate (self.grid):
            for x, cell in enumerate (row):
                if ( cell == n ):
                    print ("Tile {:d} at ({:d}, {:d})".format (n, x, y))
                    return Pos(x, y)

    def SpaceLeft (self):
        print ("SpaceLeft")
        if (self.lock[self.space.y][self.space.x - 1]):
            self.SpaceDown ()
        else:
            self.grid[self.space.y][self.space.x] = self.grid[self.space.y][self.space.x - 1]
            self.space.x -= 1
            self.grid[self.space.y][self.space.x] = 0
            self.Print ()

    def SpaceRight (self):
        print ("SpaceRight")
        self.grid[self.space.y][self.space.x] = self.grid[self.space.y][self.space.x + 1]
        self.space.x += 1
        self.grid[self.space.y][self.space.x] = 0
        self.Print ()

    def SpaceUp (self):
        print ("SpaceUp")
        if (self.lock[self.space.y - 1][self.space.x]):
            if ((self.space.x + 1 == self.tile.x) and (self.space.y == self.tile.y)):
                self.SpaceDown ()
                self.SpaceRight ()
                self.SpaceRight ()
            else:
                self.SpaceRight ()
        else:
            self.grid[self.space.y][self.space.x] = self.grid[self.space.y - 1][self.space.x]
            self.space.y -= 1
            self.grid[self.space.y][self.space.x] = 0
            self.Print ()

    def SpaceDown (self):
        print ("SpaceDown")
        self.grid[self.space.y][self.space.x] = self.grid[self.space.y + 1][self.space.x]
        self.space.y += 1
        self.grid[self.space.y][self.space.x] = 0
        self.Print ()

    def SpaceToCol (self, n):
        print ("SpaceToCol ({:d})".format (n))
        while (self.space.x < n):
            self.SpaceRight ()
        while (self.space.x > n):
            self.SpaceLeft ()

    def SpaceToRow (self, n):
        print ("SpaceToRow ({:d})".format (n))
        while (self.space.y < n):
            self.SpaceDown ()
        while (self.space.y > n):
            self.SpaceUp ()

    def TileLeft (self):
        print ("TileLeft")
        if ((self.space.y == self.tile.y) and (self.space.x > self.tile.x)):
            if (self.space.y == 3):
                self.SpaceUp ()
            else:
                self.SpaceDown ()
        self.SpaceToCol (self.tile.x - 1)
        self.SpaceToRow (self.tile.y)
        self.SpaceRight ()
        self.tile.x -= 1

    def TileRight (self):
        print ("TileRight")
        if ((self.space.y == self.tile.y) and (self.space.x < self.tile.x)):
            if (self.space.y == 3):
                self.SpaceUp ()
            else:
                self.SpaceDown ()
        self.SpaceToCol (self.tile.x + 1)
        self.SpaceToRow (self.tile.y)
        self.SpaceLeft ()
        self.tile.x += 1

    def TileUp (self):
        print ("TileUp")
        if ((self.space.x == self.tile.x) and (self.space.y > self.tile.y)):
            if (self.space.x == 3):
                self.SpaceLeft ()
            else:
                self.SpaceRight ()
        self.SpaceToRow (self.tile.y - 1)
        self.SpaceToCol (self.tile.x)
        self.SpaceDown ()
        self.tile.y -= 1

    def TileDown (self):
        print ("TileDown")
        if ((self.space.x == self.tile.x) and (self.space.y < self.tile.y)):
            if (self.space.x == 3):
                self.SpaceLeft ()
            else:
                self.SpaceRight ()
        self.SpaceToRow (self.tile.y + 1)
        self.SpaceToCol (self.tile.x)
        self.SpaceUp ()
        self.tile.y += 1

    def TileToCol (self, n):
        print ("TileToCol ({:d})".format (n))
        while (self.tile.x < n):
            self.TileRight ()
        while (self.tile.x > n):
            self.TileLeft ()

    def TileToRow (self, n):
        print ("TileToRow ({:d})".format (n))
        while (self.tile.y < n):
            self.TileDown ()
        while (self.tile.y > n):
            self.TileUp ()

    def FixupDown (self):
        print ("FixupDown")
        self.SpaceToCol (3)
        self.SpaceLeft ()
        self.SpaceUp ()
        self.SpaceRight ()
        self.SpaceDown ()
        self.SpaceDown ()
        self.SpaceLeft ()
        self.SpaceUp ()
        self.SpaceUp ()
        self.SpaceRight ()
        self.SpaceDown ()
        self.SpaceLeft ()
        self.SpaceUp ()
        self.SpaceRight ()
        self.SpaceDown

    def FixupRight (self):
        print ("FixupRight")
        self.SpaceToRow (3)
        self.SpaceUp ()
        self.SpaceLeft ()
        self.SpaceDown ()
        self.SpaceRight ()
        self.SpaceRight ()
        self.SpaceUp ()
        self.SpaceLeft ()
        self.SpaceLeft ()
        self.SpaceDown ()
        self.SpaceRight ()
        self.SpaceUp ()
        self.SpaceLeft ()
        self.SpaceDown ()
        self.SpaceRight

    def Solve (self):
        self.Print ()
        self.tile = self.Find (1)
        self.TileToCol (0)
        self.TileToRow (0)
        self.lock[0][0] = True
        self.tile = self.Find (2)
        self.TileToCol (1)
        self.TileToRow (0)
        self.lock[0][1] = True
        if ((self.grid[0][2] != 3) or (self.grid[0][3] != 4)):
            self.tile = self.Find (3)
            self.TileToCol (3)
            self.TileToRow (0)
            self.SpaceToRow (1)
            if ( self.grid[0][2] == 4 ):
                self.FixupDown ()
            self.tile = self.Find (4)
            self.TileToCol (3)
            self.TileToRow (1)
            self.SpaceToCol (2)
            self.SpaceToRow (0)
            self.SpaceRight ()
            self.SpaceDown ()
        self.lock[0][2] = True
        self.lock[0][3] = True
        self.tile = self.Find (5)
        self.TileToCol (0)
        self.TileToRow (1)
        self.lock[1][0] = True
        self.tile = self.Find (6)
        self.TileToCol (1)
        self.TileToRow (1)
        self.lock[1][1] = True
        if ((self.grid[1][2] != 7) or (self.grid[1][3] != 8)):
            self.tile = self.Find (7)
            self.TileToCol (3)
            self.TileToRow (1)
            self.SpaceToRow (2)
            if ( self.grid[1][2] == 8 ):
                self.FixupDown ()
            self.tile = self.Find (8)
            self.TileToCol (3)
            self.TileToRow (2)
            self.SpaceToCol (2)
            self.SpaceToRow (1)
            self.SpaceRight ()
            self.SpaceDown ()
        self.lock[1][2] = True
        self.lock[1][3] = True
        if ((self.grid[2][0] != 9) or (self.grid[3][0] != 13)):
            self.tile = self.Find (9)
            self.TileToRow (3)
            self.TileToCol (0)
            self.SpaceToCol (1)
            if ( self.grid[2][0] == 13 ):
                self.FixupRight ()
            self.tile = self.Find (13)
            self.TileToRow (3)
            self.TileToCol (1)
            self.SpaceToRow (2)
            self.SpaceToCol (0)
            self.SpaceDown ()
            self.SpaceRight ()
        self.lock[2][0] = True
        self.lock[3][0] = True
        if ((self.grid[2][1] != 10) or (self.grid[3][1] != 14)):
            self.tile = self.Find (10)
            self.TileToRow (3)
            self.TileToCol (1)
            self.SpaceToCol (2)
            if ( self.grid[2][1] == 14 ):
                self.FixupRight ()
            self.tile = self.Find (14)
            self.TileToRow (3)
            self.TileToCol (2)
            self.SpaceToRow (2)
            self.SpaceToCol (1)
            self.SpaceDown ()
            self.SpaceRight ()
        self.lock[2][1] = True
        self.lock[3][1] = True
        self.SpaceToRow (3)
        self.SpaceToCol (3)
        if (self.grid[2][2] == 12):
            self.SpaceUp ()
            self.SpaceLeft ()
            self.SpaceDown ()
            self.SpaceRight ()
        elif (self.grid[2][2] == 15):
            self.SpaceLeft ()
            self.SpaceUp ()
            self.SpaceRight ()
            self.SpaceDown ()

    def Check (self):
        target = [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 0]]
        for y in range (4):
            for x in range (4):
                if ( self.grid[y][x] != target[y][x] ):
                    return False
        return True

if (( bTest ) and ( len(sys.argv) <= 1 )):
    while (True):
        p = Puzzle ([])
        p.Solve ()
        if ( not p.Check () ):
            break
else:
    Puzzle(sys.argv).Solve()
